# Future wraps an asynchronous process and acts as a placeholder.
#
# A simple example:
# <tt>f = Future.eventually { call_remote_web_service }
# data = f.bind { |response| response.getContent }
# while data.pending?
#   puts "Still waiting..."
#   sleep 1
# end
# puts "Got a response: #{ data.unwrap }" </tt>
#
# Rather than explicitly waiting, you can always block on an answer:
# <tt>data = f.bind { |r| r.getContent }
# puts "Got a response: #{ data.finalize.unwrap }"</tt>

module Mon

  module Monad

    require_relative 'chainable_monad'
    require_relative 'monad'

    class Future < Monad
      include ChainableMonad

      # Create a Future, executing some function, with optional arguments.
      # Either:
      # <tt>value = Future::eventually { do_something_slow }</tt>
      # Or:
      # <tt>value = Future::eventually(myValue) { |val| do_something_slow(val) }</tt>
      def self::eventually *args, &fun
        FuturePromise::perform(fun, args)
      end

      # Wrap an object in a Future. It will start off as complete, but functions that you bind
      # to it will be asynchronous.
      # <tt>value = Future["some_username"]
      # futureBirthday = value.bind { |username| db.fetchUserInfo(username).getBirthday }  # Returns a FuturePromise, wrapping an inflight thread
      # puts "We have a birthday for some_username: #{ futureBirthday.unwrap }"</tt>
      def self::[](obj)
        FutureComplete[obj]
      end

      def eql? o
        # Time to collapse
        if o.is_a? Future
          self.unwrap == o.unwrap
        else
          self.unwrap == o
        end
      end

      def equals? o
        eql? o
      end

      def == o
        eql? o
      end

      # For use with contracts, DEPRECATED
      def self::valid?(v)
        v.is_a?(Mon::Future)
      end

      class << self
        protected :new
      end
    end

    # FuturePromise represents an asynchronous operation that is still inflight,
    # or complete.
    class FuturePromise < Future
      def initialize(thread)
        @thread = thread
      end
      
      # Equivalent to calling Future::eventually { ... }
      def self::perform(fun, args)
        if args.length == 1 and args[0].is_a? FuturePromise
          # We're waiting on something, but we don't want to block
          FuturePromise.new(Thread.new { fun.call(args[0].unwrap) })
        else
          FuturePromise.new(Thread.new(args) { |args| fun.call(*args) })
        end
      end

      # Take a block, and apply it to the value asynchronously, returing
      # a future to represent the result
      def bind &fun
        FuturePromise::perform(fun, [self])
      end

      # Block, then return the result (unwrapped)
      def unwrap
        @thread.value
      end

      # Block until the operation completes, then return the result (wrapped as a FutureComplete)
      def finalize
        value = @thread.value
        FutureComplete[(value.is_a? Future) ? value.unwrap : value]
      end

      # Are we still inflight?
      def pending?
        @thread.alive?
      end

      def to_s
        "FuturePromise[#{ @thread }]"
      end

      class << self
        protected :new
      end
    end

    # FutureComplete represents a finalized value.
    class FutureComplete < Future
      def initialize(value)
        @value = value
      end

      # You should probably be using Future[...] instead.
      def self::[](value)
        self::new(value)
      end

      # Asyrchronously apply fun to the value wrapped by this FutureComplete. Returns a FuturePromise.
      def bind &fun
        FuturePromise::perform(fun, [@value])
      end

      def pending?
        false
      end

      def unwrap
        @value
      end

      def to_s
        "FutureComplete[#{ @value }]"
      end

      class << self
        protected :new
      end
    end

  end

end
