# Lazy wraps computations which will only be performed on-demand.
#
# A simple example:
# <tt>l = Lazy.eventually(n) { perform_some_slow_operation(n) }
# data = f.bind { |result| some_even_more_costly_operation(result) }
# # At this point, no real computation has taken place
# puts "Here's the result: #{ data.unwrap } # ouch!</tt>
#
# Why would you ever want to do this? Here's an example:
# <tt>factNums = (0..100000).map { |i| Lazy[i] }.map { |m| m.bind { |i| i.factorial } }
# # We now have a list of 100000 'potential' factorial numbers, but
# # we've done this very quickly: we haven't actually done any math yet.
# fiveThousandFact = factNums[5000].unwrap
# # We've performed exactly one factorial operation! And yet we can pass around
# # and use factNums as if it really was a list of factorials. Neat!</tt>

module Mon

  module Monad

    require_relative 'chainable_monad'
    require_relative 'monad'

    # Lazy is the parent class of Pending and Final, the two states a Lazy monad can be in.
    # Use with:
    # <tt>lazyValue = Lazy[5] # Seems pointless so far...
    # lazyCalc = lazyValue.bind { |i| (0..i).map { |n| n.factorial } }.bind { |factlist| factlist.map { |i| i * i }.... # Keep right on going!
    # # We still haven't done any work!
    # puts lazyCalc.unwrap # Time to have a nap...</tt>
    # Or:
    # lazyProc = Lazy.eventually(5) { (0..5).map { |i| call_some_remote_service_ondemand(i) } }
    # # Haven't done anything yet...
    # lazyProc.sample.unwrap.map { |v| "A random response: #{ v }" } # Do one of 5 possible service calls</tt>
    class Lazy < Monad
      include ChainableMonad

      # Wrap a value in Lazy
      def self::[](obj = nil)
        if obj.is_a? Proc
          eventually(obj)
        else
          Final[obj]
        end
      end

      # Perform an operation, if necessary:
      # <tt>Lazy.eventually { 10 * 10 }</tt>
      # Or:
      # <tt>Lazy.eventually(10) { |n| n * 10 }</tt>
      def self::eventually(*args, &fun)
        Pending::eventually(fun, args)
      end

      # For contracts. Deprecated!
      def self::valid?(v)
        v.is_a? Mon::Lazy
      end

      class << self
        protected :new
      end
    end

    # Final represents a finalized lazy value (i.e. a value with no pending ops).
    class Final < Lazy
      def initialize(obj)
        @obj = obj
      end

      # You probably want Lazy[...]
      def self::[](obj = nil)
        Final.new(obj)
      end

      # Add an operation to be performed on the wrapped value, only if necessary
      def bind(&fun)
        Pending::eventually(fun, [@obj])
      end

      # Perform any pending lazy operations
      def finalize
        self
      end

      def _canBind?(name)
        @obj.respond_to? name
      end

      def _finalized?
        true
      end

      # Perform any pending lazy operations and unwrap the result
      def unwrap
        @obj
      end

      # Alias for #unwrap
      def _
        unwrap
      end

      def to_s
        "Final[#{ @obj }]"
      end

      def eql? o
        # Time to collapse
        if o.is_a? Lazy
          self.unwrap == o.unwrap
        else
          self.unwrap == o
        end
      end

      def equals? o
        eql? o
      end

      def == o
        eql? o
      end

      class << self
        protected :new
      end
    end

    # Pending represents a Lazy value with pending operations. Unwrapping or finalizing with trigger
    # said pending operations.
    class Pending < Lazy
      def initialize(fun, target = nil)
        case fun.arity
        when 1 then @fun = lambda { fun.call(target.unwrap) }
        when 0 then @fun = fun
        else raise ArgumentError.new("Bad function passed to #{ self }")
        end
      end

      # Add an operation to be performed on the wrapped value, only if necessary
      def bind(&fun)
        Pending::eventually(fun, [self])
      end

      def _canBind?(name)
        true
      end

      # Complete any pending operations and return the result, wrapped in a Final.
      # Eg: <tt>Lazy(10).bind { |i| i * i }.finalize # ==> Final[100]</tt>
      def finalize
        Final[unwrap]
      end

      # Complete any pending operations and return the result, unwrapped.
      # Eg: <tt>Lazy(10).bind { |i| i * i }.unwrap # ==> 100</tt>
      def unwrap
        @fun.call
      end

      # Alias for #unwrap
      def _
        unwrap
      end

      # Perform an operation, if necessary:
      # <tt>Lazy.eventually { 10 * 10 }</tt>
      # Or:
      # <tt>Lazy.eventually(10) { |n| n * 10 }</tt>    
      def self::eventually fun, args
        Pending.new(lambda { fun.call(*args.map { |v| (v.is_a? Lazy) ? v.unwrap : v }) })
      end

      def to_s
        "Pending[#{ @fun }]"
      end

      def eql? o
        # Time to collapse
        if o.is_a? Lazy
          self.unwrap == o.unwrap
        else
          self.unwrap == o
        end
      end

      def equals? o
        eql? o
      end

      def == o
        eql? o
      end

      class << self
        protected :new
      end
    end
  end
end
