# The List monad wraps lists. Gasp!
#
# A simple example:
# <tt>l = List[1, 2, 3]
# list9 = l.bind { |i| i * 9 } # ==> List[9, 18, 27]
# longList = l.bind { |i| [i, i * 2] } # ==> List[1, 2, 2, 4, 3, 6]</tt>
#
# Simple and straightforward.

module Mon

  module Monad

    require_relative 'chainable_monad'
    require_relative 'monad'
    
    class List < Monad

      include ChainableMonad

      def initialize(list)
        @list = list
      end

      # Create a list. Eg: List[1, 2, 3]
      def self.[](*list)
        List.new(list)
      end

      # Apply fun to all elements of the list (ala map):
      # List[1, 2, 3].map { |i| i + 5 } # ==> List[6, 7, 8]
      def bind &fun
        List.send(:new, @list.map { |i| fun.call(i) }.map { |i| (i.is_a? List) ? i.to_a : i}.flatten(1))
      end

      # Return the array wrapped by the List
      def unwrap
        to_a
      end

      # Alias for #unwrap
      def _
        to_a
      end

      def _canBind? name
        @list.all? { |i| i.respond_to?(name) }
      end

      def to_s
        (@list.length > 3) ? (tail = "...") : tail = "" 
        "List[#{ @list.take(3).join(", ") }#{ tail }]"
      end
      
      # Return the array wrapped by the list
      def to_a
        @list
      end
      
      class << self
        protected :new
      end
      
      def eql?(other)
        if other.is_a? List
          @list == other.unwrap
        elsif other.is_a? Array
          @list == other
        else
          false
        end
      end

      def ==(o)
        eql?(o)
      end
      
      def equal?(o)
        eql?(o)
      end

      def self::valid?(o)
        o.is_a?(List)
      end
      
      protected :unwrap, :_canBind?
    end
  end
end
