# The Maybe monad wraps a possible value. It allows chaining of
# operations on a potentially-null value.
#
# A simple example:
# <tt>m = Maybe[callSomeIntOrNilFunction()]
# v = (m * 3 + 2).abs
# # Depending whether the original method call returned a number or nil,
# # could be either None or Some[N].</tt>

module Mon

  module Monad

    require_relative 'chainable_monad'
    require_relative 'monad'

    # Superclass for Some and None. Can be used as follows:
    # <tt>m = Maybe[nil] # ==> None
    # m = Maybe[5]   # ==> Some[5]
    # m = Maybe[nil] * 7 # ==> None
    # m = Maybe[5] * 7 # ==> Some[35]
    # m = Maybe[call_to_fun].someOperation(3) # ==> Some[...] or None, never an error</tt>
    class Maybe < Monad
      include ChainableMonad

      # Override to catch None
      def method_missing(name, *args, &fun)
        self.bind { |o| o.send(name, *args, &fun) }
      end

      # Use to instantiate a Maybe monad:
      # <tt>m = Maybe[<either nil/false or not>]</tt>
      def self.[](obj)
        if obj.is_a? Maybe
          obj
        elsif obj
          Some[obj]
        else
          None[]
        end
      end

      # Get the value, or throw an exception (using the optional supplied msg) if it's empty
      def orFail(msg = nil)
        msg = "#{ self } is empty!" unless msg
        throw RuntimeError.new(msg)
      end

      # For Contracts, DEPRECATED
      def valid?(o)
        o.is_a? Maybe
      end

      class << self
        protected :new
      end
    end

    # The Some class represents a value that is NOT null/false
    class Some < Maybe

      def initialize(obj)
        @obj = obj
        super()
      end

      # Wrap an object. You probably want Maybe[...], but there's a slight difference:
      # Maybe[nil] # ==> None
      # Some[nil] # ==> Some[nil]
      def self.[](obj)
        if obj.is_a? None
          obj
        else
          Some.new(obj)
        end
      end

      # Unwrap the wrapped value, nil or not
      def unwrap
        @obj
      end

      def _canBind? name
        @obj.respond_to? name
      end

      # If there's a wrapped value, return it. Otherwise, either return
      # the supplied object, or execute the supplied block.
      # Eg:
      # <tt>Maybe[1].or(5) # ==> 1
      # Maybe[nil].or(5) # ==> 5
      # Maybe[nil].or { throw Exception.new("...") } # ==> Exception!
      def or obj = nil, &f
        @obj
      end

      # Get the value, or throw an exception (using the optional supplied msg) if it's empty
      def orFail(msg = nil)
        msg = "No such value!" if msg.nil?
        self::or { throw RuntimeError.new(msg) }
      end

      # If we're wrapping a value, apply fun() to it. Otherwise, None stays None.
      def bind &fun
        o = fun.call(@obj)
        if o.is_a? Maybe
          o
        elsif o
          Some[o]
        else
          None::none
        end
      end

      def to_s
        "Some[#{ @obj.to_s }]"
      end
      
      def eql?(other)
        if other.is_a? Some
          @obj == other.unwrap
        else
          @obj == other
        end
      end

      def ==(o)
        eql?(o)
      end
      
      def equal?(o)
        eql?(o)
      end
      
      class << self
        protected :new
      end

      protected :unwrap, :_canBind?
    end

    class None < Maybe

      # If we're wrapping a value, apply fun() to it. Otherwise, None stays None.
      def bind &fun
        self
      end

      # Unwrap the wrapped value, nil or not
      def unwrap
        nil
      end
      
      # If there's a wrapped value, return it. Otherwise, either return
      # the supplied object, or execute the supplied block.
      # Eg:
      # <tt>Maybe[1].or(5) # ==> 1
      # Maybe[nil].or(5) # ==> 5
      # Maybe[nil].or { throw Exception.new("...") } # ==> Exception!</tt>
      def or obj = nil, &f
        if obj and f
          f.call obj
        elsif f
          f.call
        else
          obj
        end
      end

      def _canBind? name
        true
      end

      # Return a new None object. Takes no args.
      # <tt>None[] # ==> None</tt>
      def self.[]
        None.new
      end

      def to_s
        "None"
      end
      
      def initialize
        super()
      end
      
      def eql?(other)
        other.is_a? None
      end

      def ==(o)
        eql?(o)
      end
      
      def equal?(o)
        eql?(o)
      end

      def self::none
        self.new
      end

      class << self
        protected :new
      end
      
      protected :unwrap, :_canBind?
    end
  end
end
