# The React monad wraps a changeable value, and catches changes to that value.
#
# A simple example:
# <tt>r = React[5]
# d = r * r
# # Currently r = Reactron[5], d = Reactor[25]
# r << 25
# # Now r = Reactron[25], d = Reactor[625]
# dd = d * 2
# # d = Reactor[625], dd = Reactor[1250]</tt>
#
# More or less, this is a Listener pattern. Reactrons are listenable,
# Reactors are listeners.

module Mon
  
  module Monad

    require_relative 'chainable_monad'
    require_relative 'monad'

    # The React class is the parent of Reactron and Reactor. The key difference
    # between the two is the presence of the << operator. The value of a Reactor is
    # derived from a Reactron, and cannot be directly changed. If viewed as a tree,
    # Reactrons are the root, and only the root may be changed.
    #
    # However, we can have more than one root!
    # <tt>m = React[2]  # ==> Reactron[2]
    # n = React[10] # ==> Reactron[10]
    # v = m * n     # Currently v == Reactor[20]
    # n << 5        # Now v == Reactor[10]</tt>
    #
    # Usage of React is straightforward:
    # <tt>r = React[some_value]</tt>
    class React < Monad
      include ChainableMonad
      
      # Wrap a value in a Reactron:
      # <tt>r = React["test"] # ==> Reactron["test"]
      def self::[] obj
        if (obj.is_a? Proc)
          Reactor[obj]
        else
          Reactron[obj]
        end
      end
      
      class << self
        protected :new
      end

      def self::valid?(v)
        v.is_a? React
      end
    end    

    # The Reactron class represents a changeable value, from which other
    # values can be derived (as Reactrons)
    class Reactron < React
      
      def initialize(obj)
        @obj = obj
      end

      # You should be using React[...]
      def self::[] obj
        Reactron.new(obj)
      end

      # Apply fun to the value wrapped by this Reactron. Returns a
      # Reactor. Whenever the value is changed (with <<), all derived
      # Reactors are also updated.
      def bind &fun
        Reactor[lambda { fun.call(self.unwrap) }]
      end

      def _canBind? name
        @obj.respond_to? name
      end

      # Unwrap the value contained by this Reactron
      def unwrap
        @obj
      end

      # Change the value of this Reactron
      def << obj
        @obj = obj
        self
      end

      def eql? o
        if o.is_a? React
          @obj == o.unwrap
        else
          @obj == o
        end
      end

      def equal? o
        eql? o
      end

      def ==(o)
        eql? o
      end

      def to_s
        "Reactron[#{ @obj }]"
      end
      
      class << self
        protected :new
      end

      protected :_canBind?
    end

    class Reactor < React

      def initialize(fun)
        @fun = fun
      end
      
      # Apply fun to the value wrapped by this Reactor (which in turn is
      # a transform on some Reactron), returning another
      # Reactron. Changes to the root Reactron will propagate through
      # the whole tree.
      def bind &fun
        Reactor[lambda { fun.call(self.unwrap) }]
      end

      # You want React[...]
      def self::[] fun
        Reactor.new(fun)
      end

      # Unwrap the (current) value contained by this Reactor
      def unwrap
        r = @fun.call
        r.is_a?(Reactor) ? r.unwrap : r
      end

      def _canBind? name
        unwrap.respond_to? name
      end

      def eql? o
        if o.is_a? React
          @fun.call == o.unwrap
        else
          @fun.call == o
        end
      end

      def equal? o
        eql? o
      end

      def ==(o)
        eql? o
      end

      def to_s
        "Reactor[#{ unwrap }]"
      end
      
      class << self
        protected :new
      end

      protected :_canBind?
    end
  end
end
